// TODO: consider enhancing UI to support the notions of /filter parameters/
// (that, in the case of integral parameters, would be adjustable via a slider bar)
public class VerticalContourExtractionFilter implements PixelBufferFilter2 {
	private int stippleRadius;
	private boolean directional;
	
	/**
	 * @param stippleRadius		specifies the radius of the smallest blob 
	 * 							whose contours the filter can find effectively;
	 * 							(use smaller values to detect the contours of thinner objects)
	 * @param directional		specifies whether this filter distinguishes between
	 * 							"orientation" of a contour (light-to-dark vs. dark-to-light)
	 */
	public VerticalContourExtractionFilter(int stippleRadius, boolean directional) {
		if (stippleRadius <= 0)
			throw new IllegalArgumentException("stippleRadius must be positive");
		
		this.stippleRadius = stippleRadius;
		this.directional = directional;
	}
	
	public void filter(int width, int height, int[] inPixels, int[] outPixels) {
		int off = 0;
		int edgeTopLimit = stippleRadius;
		int edgeBottom = height-stippleRadius;
		
		for (int y=0; y<edgeTopLimit; y++) {
			for (int x=0; x<width; x++, off++) {
				outPixels[off] = directional
					? Pixels.packPixel(255,128,128,128)
					: Pixels.packPixel(255,0,0,0);
			}
		}
		
		for (int y=edgeTopLimit; y<edgeBottom; y++) {
			for (int x=0; x<width; x++, off++) {
				// Take sum of the channels on the top side of the stipple's center pixel
				int sumOfRedOnFirstSide = 0;
				int sumOfGreenOnFirstSide = 0;
				int sumOfBlueOnFirstSide = 0;
				for (int i=stippleRadius, off2=off-width; i>0; i--, off2-=width) {
					int curPixel = inPixels[off2];
					sumOfRedOnFirstSide += Pixels.getRed(curPixel);
					sumOfGreenOnFirstSide += Pixels.getGreen(curPixel);
					sumOfBlueOnFirstSide += Pixels.getBlue(curPixel);
				}
				
				// Take sum of the channels on the bottom side of the stipple's center pixel
				int sumOfRedOnSecondSide = 0;
				int sumOfGreenOnSecondSide = 0;
				int sumOfBlueOnSecondSide = 0;
				for (int i=stippleRadius, off2=off+width; i>0; i--, off2+=width) {
					int curPixel = inPixels[off2];
					sumOfRedOnSecondSide += Pixels.getRed(curPixel);
					sumOfGreenOnSecondSide += Pixels.getGreen(curPixel);
					sumOfBlueOnSecondSide += Pixels.getBlue(curPixel);
				}
				
				// Calculate the average (per pixel) difference between the channels 
				// for the pixels on either side of the stipple's center pixel
				int avgDeltaOfRed = (sumOfRedOnSecondSide - sumOfRedOnFirstSide)/stippleRadius;
				int avgDeltaOfGreen = (sumOfGreenOnSecondSide - sumOfGreenOnFirstSide)/stippleRadius;
				int avgDeltaOfBlue = (sumOfBlueOnSecondSide - sumOfBlueOnFirstSide)/stippleRadius;
				
				// Normalize deltas, which are in the range [-255, 255], to be in the range [0,255]
				// (The method of normalization depends on whether this filter is directional or not.)
				int normalizedDeltaOfRed, normalizedDeltaOfGreen, normalizedDeltaOfBlue;
				if (directional) {
					normalizedDeltaOfRed = 128 + avgDeltaOfRed/2;
					normalizedDeltaOfGreen = 128 + avgDeltaOfGreen/2;
					normalizedDeltaOfBlue = 128 + avgDeltaOfBlue/2;
				} else {
					normalizedDeltaOfRed = Math.abs(avgDeltaOfRed);
					normalizedDeltaOfGreen = Math.abs(avgDeltaOfGreen);
					normalizedDeltaOfBlue = Math.abs(avgDeltaOfBlue);
				}
				
				outPixels[off] = Pixels.packPixel(
					255,							// alpha
					normalizedDeltaOfRed,			// red
					normalizedDeltaOfGreen,			// green
					normalizedDeltaOfBlue			// blue
				);
			}
		}
		
		for (int y=edgeBottom; y<height; y++) {
			for (int x=0; x<width; x++, off++) {
				outPixels[off] = directional
					? Pixels.packPixel(255,128,128,128)
					: Pixels.packPixel(255,0,0,0);
			}
		}
	}
	
	public String toString() {
		return "Contour Extraction (Vertical)";
	}
}